/*-
 * Copyright (C)2020 @BABOLO http://www.babolo.ru/
 * Inspired by su2-1.3.
 * This version written from scratch and FreeBSD's /usr/bin/su
 *
 * Copyright (c) 2002, 2005 Networks Associates Technologies, Inc.
 * All rights reserved.
 *
 * Portions of this software were developed for the FreeBSD Project by
 * ThinkSec AS and NAI Labs, the Security Research Division of Network
 * Associates, Inc.  under DARPA/SPAWAR contract N66001-01-C-8035
 * ("CBOSS"), as part of the DARPA CHATS research program.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright (c) 1988, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2020 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: su2.c,v 1.16 2020/10/04 20:39:17 babolo Exp $"

#define BLIN_COMPAT  3
#define Bpars_COMPAT 4
#define MIFE_COMPAT  5

#include <sys/param.h>
#include <login_cap.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <pwd.h>
#include <err.h>
#include <security/pam_appl.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "su2.h"
#include "su2_env.h"

extern char **environ;
static char  *cleanenv = NULL;

static void
usage(int ex) {
    fprintf( stderr
           , "su2 @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "     Switch user rights\n"
             "Usage: su2 [-cf] [-u fromUID] [-nice] [toUID] [-c command | -- prog args...]\n"
             "    -- ...      - execve(...)\n"
             "    -           - login\n"
             "    -c command  - exec 'sh -c command'\n"
             "    -d          - run /bin/csh\n"
             "    -h          - \n"
             "    -n nice     - nice\n"
             "    -q          - quiet\n"
             "    -s          - register new UID in utmp\n"
             "    -u fromUID  - whom rights used before user switsh\n"
             "    -v[v[v...]] - verbose\n"
           )
    ;
    exit(ex);
}

int
main(int argc, char *argv[]) {
    int                  statusp;
    struct sigaction     sa_pipe;
    struct sigaction     sa_quit;
    struct sigaction     sa_int;
    char               **envpam;
    pid_t                gchild;
    pid_t                child;
    const char          *flar    = "-a:c:dln:qst:u:v=:?";
    int                  temp;
    int                  fds[2];
    pid_t                pid;
    babolo_opts         *bos     = NULL;
    su2_cf              *cf      = NULL;
    struct sigaction     sa;
    login_cap_t         *lc;
    int                  ex      = EX_OK;
    const char          *o;
    int                  c;

    if  (!(cf = calloc(1, sizeof(*cf)))) {
        warnx("Malloc failed #1");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
#   define blin_internal_flags (cf->flags)
    openlog("su2", LOG_CONS, LOG_AUTH);
    BLIN_VERBOSE(cf->flags);
    if  (!(bos = babolo_openopts(Bpars_NONU | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    for (; ;) {
        c = babolo_getopts(bos);
        ifBLIN_QX4("-%c(%02X)%02X", c ? c : 1, c, bos->c);
        if  (!c) break;
        if  (0 > c) {
            ifBLIN_QW0("babolo_getopts -%c", bos->c);
            ex = EX_USAGE;
            goto out;
        }
        switch (c) {
        case '-':
        case 'l': cf->flags |= SU2_LOGIN;
                ; break;
        case 'a': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (!!cf->frompw) {
                ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cf->frompw->pw_name, o);
                ; }
                ; if  (!!su2_iniud(cf, bos->c, o, &cf->frompw, 0)) {
                ;     ifBLIN_QW0("No %s fromUID", o);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'c': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (!!cf->command) {
                ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cf->command, o);
                ; }
                ; if  (!(cf->command = strdup(o))) {
                ;     ifBLIN_QW0("strdup", o);
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'd': cf->flags |= SU2_CSH;
                ; break;
        case 'n': {   int n;

                ;     if  (SU2_NICE & cf->flags) {
                ;         cf->flags |= SU2_RENICE;
                ;         n = cf->nice;
                ;     }
                ;     cf->flags |= SU2_NICE;
                ;     errno = 0;
                ;     cf->nice = babolo_getoptunum(bos, 0);
                ;     if  (!!errno) {
                ;         ifBLIN_QW0("babolo_getoptunum");
                ;         ex = EX_USAGE;
                ;         goto out;
                ;     }
                ;     if  (SU2_RENICE & cf->flags) {
                ;         ifBLIN_QX1("Redefined -%c option: %d->%d", bos->c, n, cf->nice);
                ; }   }
                ; break;
        case 'q': BLIN_QUIET(cf->flags);
                ; break;
        case 's': cf->flags |= SU2_UREG;
                ; break;
        case 't': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (!!cf->class) {
                ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cf->class, o);
                ; }
                ; cf->class = optarg;
                ; break;
        case 'u': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (!!cf->topw) {
                ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cf->topw->pw_name, o);
                ; }
                ; if  (!!su2_iniud(cf, bos->c, o, &cf->topw, 0)) {
                ;     ifBLIN_QW0("No %s fromUID", o);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'v': BLIN_VERBOSE(cf->flags);
                ; break;
        case '=': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; switch(o[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);    exit(EX_OK);
                  case '1': babolo_optest(flar);        exit(EX_OK);
                  case '2': babolo_optext(bos);         exit(EX_OK);
                  case '3': babolo_dumpopts(bos);       exit(EX_OK);
                  case 'D': cf->flags |= SU2_DUMP; break;
                  default : usage(EX_USAGE);
                  }
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", c);
                ; usage(EX_USAGE);
    }   }
    ifBLIN_QX4("%08X %08X", bos->flags, Bpars_DASH);
    if  (Bpars_DASH & bos->flags) cf->flags |= SU2_DASH;
    cf->argp = cf->amax = ARGSHIFT;
    for (int i = cf->argp; i < MAXARGS;) {
        if  (!(o = babolo_getargs(bos))) break;
        if  (!(cf->args[i] = strdup(o))) {
            ifBLIN_QW0("strdup");
            ex = EX_OSERR;
            goto out;
        }
        cf->amax = ++i;
    }
    if  (!!geteuid()) {
        ifBLIN_QX0("Not running setuid root");
        ex = EX_USAGE;
#if 0
        goto out;
#endif
    }
    cf->uid = getuid();
    cf->gid = getgid();
    if  (!!su2_iniud(cf, ':', getlogin(), &cf->mypw, cf->uid)) {
        ifBLIN_QW0("No my UID %d", cf->uid);
        ex = EX_USAGE;
        goto out;
    }
    if  (!cf->frompw && !!su2_iniud(cf, '-', NULL, &cf->frompw, cf->uid)) {
        ifBLIN_QW0("No fromUID");
        ex = EX_OSERR;
        goto out;
    }
    if  (!(SU2_DASH & cf->flags)) {
        ifBLIN_QX4("DASH 0=%s~", !cf->args[ARGSHIFT] ? "" : cf->args[ARGSHIFT]);
        if  (!cf->topw && !!su2_iniud(cf, '-', cf->args[ARGSHIFT], &cf->topw, 0)) {
            ifBLIN_QW0("No %s toUID", !cf->args[ARGSHIFT] ? "root" : cf->args[ARGSHIFT]);
            ex = EX_USAGE;
            goto out;
        }
        if  (!!cf->args[ARGSHIFT]) ++cf->argp;
        if  (!!cf->args[ARGSHIFT + 1]) {
        ;   ifBLIN_QX4("1=%s~", cf->args[ARGSHIFT + 1]);
        ;   if  (!strncmp(cf->args[ARGSHIFT + 1], "-c", 2)) {
        ;       ifBLIN_QX4("-c");
        ;       if  (2 < strnlen(cf->args[ARGSHIFT + 1], 3)) {
        ;           if  (!!cf->command) {
        ;               ifBLIN_QX1( "Command redefined from=%s~ to=%s~"
                                  , cf->command
                                  , &cf->args[ARGSHIFT + 1][2]
                                  );
        ;           }
        ;           if  (!(cf->command = strdup(&cf->args[ARGSHIFT + 1][2]))) {
        ;               ifBLIN_QW0("strdup", &cf->args[ARGSHIFT + 1][2]);
        ;               ex = EX_OSERR;
        ;               goto out;
        ;           }
        ;       } else if (!!cf->args[ARGSHIFT + 2]) {
        ;           ifBLIN_QX4("2=%s~", cf->args[ARGSHIFT + 2]);
        ;           if  (!!cf->command) {
        ;               ifBLIN_QX1( "Command redefined from=%s~ to=%s~"
                                  , cf->command
                                  , cf->args[ARGSHIFT + 2]
                                  );
        ;           }
        ;           if  (!(cf->command = strdup(cf->args[ARGSHIFT + 2]))) {
        ;               ifBLIN_QW0("strdup", cf->args[ARGSHIFT + 2]);
        ;               ex = EX_OSERR;
        ;               goto out;
        ;           }
        ;       } else {
        ;           ifBLIN_QX0("Command omited");
        ;           ex = EX_USAGE;
        ;           goto out;
        ;       }
        ;   } else if (!!strcmp(cf->args[ARGSHIFT + 1], "--")) {
        ;       ifBLIN_QX0("No --");
        ;       ex = EX_USAGE;
        ;       goto out;
        ;   } else {
        ;       cf->flags |= SU2_DASH;
        ;       ++cf->argp;
        }   }
    } else if (!cf->topw && !!su2_iniud(cf, ':', "root", &cf->topw, 0)) {
        ifBLIN_QW0("No root toUID");
        ex = EX_USAGE;
        goto out;
    }
    endpwent();
    if  (!!(ex = su2_shell(cf))) {
        ifBLIN_QX0("su2_shell");
        goto out;
    }
    if  (!cf->frompw) {
        ifBLIN_QX0("Unknown fromUID");
        ex = EX_USAGE;
        goto out;
    }
    if  (!cf->topw) {
        ifBLIN_QX0("Unknown toUID");
        ex = EX_USAGE;
        goto out;
    }
    if  (!!(ex = su2_nice(cf))) {
        ifBLIN_QX0("su2_nice");
        goto out;
    }
    if  (!!(ex = su2_tty(cf))) {
        ifBLIN_QX0("su2_tty");
        goto out;
    }
    if  (SU2_DUMP & cf->flags) su2_dump(cf);
    if  (!!(ex = su2_rc(cf))) {
        ifBLIN_QX0("su2_rc");
        goto out;
    }
    if  (!((SU2_NOCHCK | SU2_UFOUND) & (cf->cond | cf->next))) {
        su2_log(cf, 0);
        ifBLIN_QX0("Denied");
        ex = EX_NOPERM;
        goto out;
    }
    if  (!!(ex = su2_auth(cf))) {
        ifBLIN_QX0("Not preauthorized");
        ex = EX_NOPERM;
        goto out;
    }
    if  (  !!cf->uid
        && (  (cf->uid != cf->frompw->pw_uid)
           || (~(cf->next | cf->cond) & SU2_NOPASS)
           )
        && !!(ex = su2_inpass(cf))
        ) {
        su2_log(cf, 0);
        ifBLIN_QX0("Not authorized");
        ex = EX_NOPERM;
        goto out;
    }
    su2_log(cf, 1);
    if  (!!(ex = su2_reg(cf, 1))) {
        ifBLIN_QW0("su2_reg in");
        goto out;
    }
    if  (!cf->class) {
        if  (!(lc = login_getpwclass(cf->topw))) {
            ifBLIN_QW0("login_getpwclass");
            ex = EX_OSERR;
            goto out;
        }
    } else if (!cf->mypw->pw_uid && !cf->topw->pw_uid) {
        ifBLIN_QX0("Only root may use -c");
        ex = EX_NOPERM;
        goto out;
    } else if (!(lc = login_getclass(cf->class))) {
        ifBLIN_QW0("login_getclass %s", cf->class);
        ex = EX_OSERR;
        goto out;
    }
    if  (!!cf->class && (!lc->lc_class || !!strcmp(cf->class, lc->lc_class))) {
        ifBLIN_QX0("Unknown class: %s", cf->class);
        ex = EX_USAGE;
        goto out;
    }
    /* PAM modules might add supplementary groups in pam_setcred(), so initialize them first. */
    if  (0 > setusercontext(lc, cf->topw, cf->topw->pw_uid, LOGIN_SETGROUP)) {
        ifBLIN_QW0("setusercontext");
        ex = EX_OSERR;
        goto out;
    }
    if  (PAM_SUCCESS != (cf->pamret = pam_setcred(cf->pamh, PAM_ESTABLISH_CRED))) {
        ifBLIN_QX0("pam_setcred: %s", pam_strerror(cf->pamh, cf->pamret));
        ex = EX_OSERR;
        goto out;
    }
    if  (  (SU2_LOGIN & cf->flags)
        && (PAM_SUCCESS != (cf->pamret = pam_open_session(cf->pamh, 0)))
        ) {
        ifBLIN_QX0("pam_open_session: %s", pam_strerror(cf->pamh, cf->pamret));
        ex = EX_OSERR;
        goto out;
    }
    sa.sa_flags = SA_RESTART;
    sa.sa_handler = SIG_IGN;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGINT, &sa, &sa_int);
    sigaction(SIGQUIT, &sa, &sa_quit);
    sigaction(SIGPIPE, &sa, &sa_pipe);
    sa.sa_handler = SIG_DFL;
    sigaction(SIGTSTP, &sa, NULL);
    statusp = 1;
    if  (pipe(fds) == -1) {
        ifBLIN_QX0("pipe");
        ex = EX_IOERR;
        if  (!!su2_pamend(cf)) ifBLIN_QX1("su2_pamend");
        goto out;
    }
    switch ((child = fork())) {
    case -1:
        ifBLIN_QW0("fork");
        ex = EX_OSERR;
        if  (!!su2_pamend(cf)) ifBLIN_QX1("su2_pamend");
        goto out;
    case 0 :
        close(fds[1]);
        read(fds[0], &temp, 1);
        close(fds[0]);
        sigaction(SIGPIPE, &sa_pipe, NULL);
        sigaction(SIGINT, &sa_int, NULL);
        sigaction(SIGQUIT, &sa_quit, NULL);
        if  (0 > setusercontext( lc
                               , cf->topw
                               , cf->topw->pw_uid
                               , LOGIN_SETUSER | LOGIN_SETCPUMASK | LOGIN_SETLOGINCLASS
            )                  ) {
            ifBLIN_QW0("setusercontext(SETUSER SETCPUMASK SETLOGINCLASS)");
            ex = EX_OSERR;
            goto out;
        }
        if  (SU2_LOGIN & cf->flags) {
            const char    *p;
            char         **e;
            char          *v;

            p = getenv("TERM");
            environ = &cleanenv;
            if  (!!setenv("HOME", cf->topw->pw_dir, 1)) {
                ifBLIN_QW1("setenv(HOME, %s, 1)", cf->topw->pw_dir);
            }
            if  (!!setenv("SHELL", cf->shell, 1)) ifBLIN_QW1("setenv(SHELL, %s, 1)", cf->shell);
            if  (!!(envpam = pam_getenvlist(cf->pamh))) {
                for (e = envpam; !!*e; ++e) {
                    if  (1024 < strlen(*e)) continue;
                    if  (!strncmp(*e, "LD_", 3)) continue;
                    if  (!(v = strchr(*e, '='))) continue;
                    *(v++) = '\0';
                    if  (1 == babolo_testword(&env, *e)) continue;
                    if  (!!setenv(*e, v, 1)) ifBLIN_QW1("setenv(setenv(%s, %s, 1)", *e, v);
            }   }
            if  (0 > setusercontext( lc
                                   , cf->topw
                                   , cf->topw->pw_uid
                                   , LOGIN_SETPATH | LOGIN_SETUMASK | LOGIN_SETENV | LOGIN_SETRESOURCES
                )                  ) {
                ifBLIN_QW0("setusercontext(LOGIN_SETPATH LOGIN_SETUMASK LOGIN_SETENV)");
                ex = EX_OSERR;
                goto out;
            }
            if  (!!p && !!setenv("TERM", p, 1)) ifBLIN_QW1("setenv(TERM, %s, 1)", p);
            p = pam_getenv(cf->pamh, "HOME");
            if  (!!chdir(p ? p : cf->topw->pw_dir)) {
                ifBLIN_QW0("chdir %s", p ? p : cf->topw->pw_dir);
                ex = EX_UNAVAILABLE;
                goto out;
        }   }
        login_close(lc);
        ifBLIN_QX3( "to execv %s(%s, %s, %s, %s)"
                  , (SU2_DASH & cf->flags) ? cf->args[cf->argp] : cf->shell
                  , ((cf->argp + 1) < cf->amax) ? cf->args[cf->argp + 1] : ""
                  , ((cf->argp + 2) < cf->amax) ? cf->args[cf->argp + 2] : ""
                  , ((cf->argp + 3) < cf->amax) ? cf->args[cf->argp + 3] : ""
                  , ((cf->argp + 4) < cf->amax) ? cf->args[cf->argp + 4] : ""
                  );
        execv((SU2_DASH & cf->flags) ? cf->args[cf->argp] : cf->shell, &cf->args[cf->argp]);
        ifBLIN_QX0("exec");
        ex = EX_OSERR;
        goto out;
    default:
        sa.sa_handler = SIG_IGN;
        sigaction(SIGTTOU, &sa, NULL);
        close(fds[0]);
        setpgid(child, child);
        if  (!!cf->tty && (tcgetpgrp(cf->ttyfd) == getpgrp())) tcsetpgrp(cf->ttyfd, child);
        close(fds[1]);
        sigaction(SIGPIPE, &sa_pipe, NULL);
        while (-1 != (pid = waitpid(child, &statusp, WUNTRACED))) {
            if  (WIFSTOPPED(statusp)) {
                gchild = getpgid(child);
                if  (!!cf->tty && (tcgetpgrp(cf->ttyfd) == gchild)) {
                    tcsetpgrp(cf->ttyfd, getpgrp());
                }
                kill(getpid(), SIGSTOP);
                if  (!!cf->tty && (tcgetpgrp(cf->ttyfd) == getpgrp())) {
                    gchild = getpgid(child);
                    tcsetpgrp(cf->ttyfd, gchild);
                }
                kill(child, SIGCONT);
                statusp = 1;
                continue;
            }
            break;
        }
        if  (!!cf->tty) tcsetpgrp(cf->ttyfd, getpgrp());
        if  (pid == -1) err(1, "waitpid");
        if  (!!su2_pamend(cf)) ifBLIN_QX1("su2_pamend");
        ex = WEXITSTATUS(statusp);
        if  (!!(ex = su2_reg(cf, 0))) {
            ifBLIN_QW0("su2_reg out");
            goto out;
    }   }
out:
    exit(ex);
#   undef blin_internal_flags
}
